/*
* SPDX-License-Identifier: Apache-2.0
* SPDX-FileCopyrightText: 2013 German Aerospace Center (DLR)
*
* Created: 2014-02-07 Martin Siggel <Martin.Siggel@dlr.de>
*/

#ifndef GORDON_INTERNAL_H
#define GORDON_INTERNAL_H

#define VERSION_HEX_CODE(MAJOR, MINOR, PATCH) ((MAJOR) << 16 | (MINOR) << 8 | (PATCH))

#endif // GORDON_INTERNAL_H
